import java.util.LinkedList;

import Jcg.geometry.Point_;
import Jcg.geometry.Point_2;
import Jcg.geometry.Point_3;
import Jcg.geometry.Point_d;

/**
 * A class providing basic tools for dealing with point sets (in 2D, 3D or arbitrary dimension)
 * 
 * @author Luca Castelli Aleardi (Ecole Polytechnique, INF562)
 * @version jan 2020
 */    
public class PointSetUtil {

	/**
	 * Split a point set into two point sub-sets, cutting around the 'median' point in a given direction (lower an upper point sub-sets). <br>
	 * If there are more points having the same d-coordinate coinciding with the 'cutValue', then we split in order to get a balanced partition.
	 * 
	 * @param pointSet  the input point set to be split
	 * 
	 * @return an array of size 2, storing the two sub-sets (points are stored using linked lists)
	 */    
	public static<X extends Point_> LinkedList<X>[] split(LinkedList<X> pointSet, double cutValue, int cutDim) {
		LinkedList<X> medianPoints=new LinkedList<X>(); // points whose coordinate coincides with the cutValue
		
		LinkedList<X>[] subN=new LinkedList[2]; // subsets of points, resulting from the splitting around the median 
		subN[0] = new LinkedList<X>(); 
		subN[1] = new LinkedList<X>();
		
		int equal=0;
		int m0=0; // number of elements 'p' inserted in the first list, such that p.Cartesian(dim)=cutValue
		int m1=0; // number of elements 'p' inserted in the second list, such that p.Cartesian(dim)=cutValue
		for (X p: pointSet) {
			if((p.getCartesian(cutDim).doubleValue() < cutValue))
				subN[0].add(p);
			else if((p.getCartesian(cutDim).doubleValue() > cutValue))
				subN[1].add(p);
			else { // we have to deal with the case where there are more points having the same d-coordinates corresponding to the 'cutValue', in order to get a balanced partition
				medianPoints.add(p); // store these points in a temporary array
			}
		}
		
		int n1=subN[0].size();
		//int n2=subN[1].size();
		int N=pointSet.size()/2-n1;
		
		int counter=0;
		for(X p: medianPoints) {
			if(counter<N)
				subN[0].add(p);
			else
				subN[1].add(p);
			
			counter++;
		}
		
		return subN;
	}

	/**
	 * Compute the bounding box of a point set
	 * 
	 * @param pointSet  the input point set
	 * 
	 * @return an array of size 2*dim, storing the coordinates of the bounding box [xmin, ymin, zmin, ..., xmax, ymax, zmax, ...]
	 */    
	public static<X extends Point_> double[] boundingBox(LinkedList<X> points) {
		if(points==null || points.size()==0) {
			System.out.println("point cloud empty");
			return null;
		}
		int dim=points.getFirst().dimension();
		double[] box=new double[2*dim];

		for(int i=0;i<dim;i++) {
			box[i]=Double.MAX_VALUE;
			box[dim+i]=Double.MIN_VALUE;
		}

		for(X p: points) {
			for(int i=0;i<dim;i++) {
				// updating minimum values
				if(p.getCartesian(i).doubleValue()<box[i])
					box[i]=p.getCartesian(i).doubleValue();
				// updating maximum values
				if(p.getCartesian(i).doubleValue()>box[i+dim])
					box[i+dim]=p.getCartesian(i).doubleValue();
			}
		}
		return box;
	}
    
	   /**
	   * Compute the mean (barycenter) of a point set
	   */
	  public static<X extends Point_> X mean(LinkedList<X> points) {
		  if(points==null || points.size()==0)
			  return null;

		  int dim = points.getFirst().dimension();
		  double[] coords = new double [dim];
		  int totalWeight = 0;

		  for (X p: points) {
			  totalWeight++;
			  for (int i=0; i<dim; i++)
				  coords[i] += p.getCartesian(i).doubleValue();
		  }
		  for (int i=0; i<dim; i++)
			  coords[i] /= totalWeight;
		  
		  Point_ result=null;
		  if(dim==2)
			  result=new Point_2 (coords[0], coords[1]);
		  else if(dim==3)
			  result=new Point_3 (coords[0], coords[1], coords[2]);
		  if(dim>3)
			  result=new Point_d (coords);
		  
		  return (X)result;
	  }

    /**
     * return a point cloud of 'n' random points sampled on a 2D annulus
     * (points are sampled according to normal distribution)
     */
    public static LinkedList<Point_2> randomPointsOnAnnulus(int n, double radius) {
    	System.out.print("Generating 2D point set in disk...");
    	LinkedList<Point_2> result=new LinkedList<Point_2>();
    	double R=1.;
    	int dim=2;
    	for (int i=0; i<n; i++) {
    	    double[] c = new double[dim];
    	    double angle=Math.random()*2*Math.PI;
    	    for (int j=0; j<dim; ++j) {    	
    	    	c[j] = -radius+radius*2.*Math.random();
    	    }
    	    c[0]=c[0]+R*Math.cos(angle);
    	    c[1]=c[1]+R*Math.sin(angle);
    	    result.add(new Point_2 (c[0], c[1]));
    	}
    	System.out.println("done ("+n+" points)");
    	return result;
    }

    /**
     * return a point cloud of n random points (in the unit 2D square)
     */
    public static LinkedList<Point_2> randomPointsInUnitSquare(int n) {
    	int dim=2;
    	LinkedList<Point_2> result=new LinkedList<Point_2>();
    	for (int i=0; i<n; i++) {
    	    double[] c = new double[dim];
    	    for (int j=0; j<dim; ++j)
    	    	c[j] = 1.*Math.random();
    	    result.add(new Point_2 (c[0], c[1]));
    	}
    	System.out.println("Generated 2D point set in unit square ("+n+" points)");
    	return result;
    }

    /**
     * return a point cloud of n random points (in the unit 3D cube)
     */
    public static LinkedList<Point_3> randomPointsInUnitCube(int n) {
    	int dim=3;
    	LinkedList<Point_3> result=new LinkedList<Point_3>();
    	for (int i=0; i<n; i++) {
    	    double[] c = new double[dim];
    	    for (int j=0; j<dim; ++j)
    	    	c[j] = 1.*Math.random();
    	    result.add(new Point_3 (c[0], c[1], c[2]));
    	}
    	System.out.println("Generated 3D point set in unit cube ("+n+" points)");
    	return result;
    }

    /**
     * return a point cloud of n random points sampled on a circle 
     * (according to normal distribution)
     */
    public static<X> void print(LinkedList<X> points) {
    	for(X p: points)
    		System.out.println(" "+p);
    	System.out.println();
    }

}
