import java.util.Calendar;
import java.util.LinkedList;

import Jcg.geometry.Point_;
import Jcg.geometry.Point_2;
import Jcg.geometry.Point_3;
import Jcg.geometry.Point_d;

/**
 * Class for testing and benchmarking range search methods
 * 
 * @author Luca Castelli Aleardi (Ecole Polytechnique, INF562)
 * @version jan 2014
 */
public class TestRangeSearch {
	
	static long queryTime=0;
	static long constructionTime=0;
	static long basicQueryTime=0;    

	/**
	 * Show time performances
	 */
	public static void showTimePerformances() {
		System.out.println("Range Search performaces:\nTotal construction time: " + constructionTime/1000 + "s " + 
				constructionTime%1000 + "ms");   	
		System.out.println("Total NN query time: " + queryTime/1000 + "s " + 
				queryTime%1000 + "ms\n");   	
	}

	/**
	 * Main test function, for testing 2D and 3D range search
	 */
	public static void main(String args[]) {
		System.out.println("Exo 1.3: testing Range Search");
		LinkedList points2D=null, points3D=null; // input point cloud
		double radius; // radius of the query ball
		LinkedList<Point_2> query2D=null;
		Point_3 query3D=null;
		int nTests=10000;
		int nPoints=200000;
		
		if (args.length < 1) {
			points2D=(LinkedList<Point_2>)PointSetUtil.randomPointsOnAnnulus(nPoints, 0.1);
			query2D=(LinkedList<Point_2>)PointSetUtil.randomPointsInUnitSquare(nTests);
			radius=0.01;
			
			Draw.draw2D(points2D, "original point cloud");
			
			System.out.print("Initializing Kd-Tree data structure... ");
			RangeSearch<Point_2> fastSearch=new FastRangeSearch<Point_2>(points2D);
			System.out.println(" done");
			RangeSearch<Point_2> slowSearch=new SlowRangeSearch<Point_2>(points2D);
			
			// running test for checking the result
			LinkedList<Point_2> result=checkRangeQueries(slowSearch, fastSearch, query2D.get(0), 10*radius);
			if(result!=null && result.size()>0)
				Draw.draw2D(result, "query result (via Kd-Trees)");

			// Running benchmarks for evaluation time complexity
			System.out.println();
			runBenchmarks(slowSearch, query2D, radius);
			runBenchmarks(fastSearch, query2D, radius);

			System.out.println("test done.");
			FastRangeSearch.timePerformance();
		}
		
	}

	/**
	 * Benchmarks for evaluating and comparing running times
	 * 
	 * @param search	the data structure to be used to perform queries
	 * @param queries	set of query points
	 * @param radius	size of the query window
	 */
	public static<X extends Point_> void runBenchmarks(RangeSearch<X> search, LinkedList<X> queries, double radius) {
		//System.out.println("Performing range queries: running benchmarks");
		Calendar rightNow = Calendar.getInstance(); // to compute query time
		long start = rightNow.getTimeInMillis();
		
		System.out.print("Running benchmark for "+search.toString()+"...");

		LinkedList<X> neighbors;
		for(X q: queries) {
			neighbors=search.OrthogonalRangeSearch(q, radius);
		}
		
		rightNow = Calendar.getInstance(); // to compute query time
		long end = rightNow.getTimeInMillis();
		
		double n=queries.size();
		double timePerQuery=(end-start)/n;
		System.out.println("done");
		System.out.println("\t elapsed time: "+(end-start)+" ms (time per query: "+timePerQuery+" ms)");
		System.out.println("\t number of query points: "+queries.size());
	}


	/**
	 * Test for evaluating the correctness of the results
	 * 
	 * @param searchSlow	slow data structure for range queries
	 * @param searchFast	fast data structure for range queries
	 * @param q			query point
	 * @param radius	size of the query window
	 */
	public static<X extends Point_> LinkedList<X> checkRangeQueries(RangeSearch<X> searchSlow, RangeSearch<X> searchFast, X q, double radius) {
		System.out.print("Checking correctness: \""+searchSlow.toString()+"\" versus \""+searchFast.toString()+"\"...");

		LinkedList<X> neighbors1=searchSlow.OrthogonalRangeSearch(q, radius);
		LinkedList<X> neighbors2=searchFast.OrthogonalRangeSearch(q, radius);

		if(neighbors1.size()!=neighbors2.size())
			throw new Error("error: wrong number of neighbors: "+neighbors1.size()+", "+neighbors2.size());
		System.out.println("ok ("+neighbors1.size()+" neighbors)");
		
		return neighbors2;
	}

}
